from typing import Any, Dict, Iterable, Optional, Set, BinaryIO, Union

from cchardet import UniversalDetector
import networkx as nx
import numpy as np
import pandas as pd
from pandas.core.common import flatten


def setwrap(value: Any) -> Set[str]:
    """
    Returns a flattened and stringified set from the given object or iterable.

    For use in public functions which accept argmuents or kwargs that can be
    one object or a list of objects.
    """
    return set(map(str, set(flatten([value]))))


def remove_node_attributes(G: nx.DiGraph, attributes: Union[str, Iterable[str]]):
    """
    Return a copy of the graph with the given attributes
    deleted from all nodes.
    """
    G = G.copy()
    for _, data in G.nodes(data=True):
        for attribute in setwrap(attributes):
            if attribute in data:
                del data[attribute]
    return G


def detect_encoding(f: BinaryIO, limit: int = 2500) -> str:
    """
    Return encoding of provided input stream.

    Most of the time it's unicode, but if we are unable to decode the input
    natively, use `chardet` to determine the encoding heuristically.
    """
    unicode_decodable = True

    for line_no, line in enumerate(f):
        try:
            line.decode("utf-8")
        except UnicodeDecodeError:
            unicode_decodable = False
            break

        if line_no > limit:
            break

    if unicode_decodable:
        return "utf-8"

    f.seek(0)
    u = UniversalDetector()

    for line_no, line in enumerate(f):
        u.feed(line)
        if u.done or line_no > limit:
            break

    u.close()
    return u.result["encoding"]


def empty_df(columns: Optional[Iterable[str]] = None) -> pd.DataFrame:
    columns = [] if columns is None else columns
    empty: Dict = {col: [] for col in columns}
    return pd.DataFrame(empty, columns=columns, dtype=np.unicode)



import os
from threading import RLock
from typing import Dict, Optional, Union

import networkx as nx
import numpy as np
import pandas as pd

from .config import default_config
from .types import View
from .utilities import detect_encoding, empty_df, setwrap


def _read_file(filename: str) -> property:
    def getter(self) -> pd.DataFrame:
        return self.get(filename)

    return property(getter)


class Feed(object):
    def __init__(
        self,
        source: Union[str, "Feed"],
        view: Optional[View] = None,
        config: Optional[nx.DiGraph] = None,
    ):
        self._config: nx.DiGraph = default_config() if config is None else config
        self._view: View = {} if view is None else view
        self._cache: Dict[str, pd.DataFrame] = {}
        self._pathmap: Dict[str, str] = {}
        self._delete_after_reading: bool = False
        self._shared_lock = RLock()
        self._locks: Dict[str, RLock] = {}
        if isinstance(source, self.__class__):
            self._read = source.get
        elif isinstance(source, str) and os.path.isdir(source):
            self._read = self._read_csv
            self._bootstrap(source)
        else:
            raise ValueError("Invalid source")

    def get(self, filename: str) -> pd.DataFrame:
        lock = self._locks.get(filename, self._shared_lock)
        with lock:
            df = self._cache.get(filename)
            if df is None:
                df = self._read(filename)
                df = self._filter(filename, df)
                df = self._prune(filename, df)
                self._convert_types(filename, df)
                df = df.reset_index(drop=True)
                df = self._transform(filename, df)
                self.set(filename, df)
            return self._cache[filename]

    def set(self, filename: str, df: pd.DataFrame) -> None:
        lock = self._locks.get(filename, self._shared_lock)
        with lock:
            self._cache[filename] = df

    agency = _read_file("agency.txt")
    calendar = _read_file("calendar.txt")
    calendar_dates = _read_file("calendar_dates.txt")
    fare_attributes = _read_file("fare_attributes.txt")
    fare_rules = _read_file("fare_rules.txt")
    feed_info = _read_file("feed_info.txt")
    frequencies = _read_file("frequencies.txt")
    routes = _read_file("routes.txt")
    shapes = _read_file("shapes.txt")
    stops = _read_file("stops.txt")
    stop_times = _read_file("stop_times.txt")
    transfers = _read_file("transfers.txt")
    trips = _read_file("trips.txt")

    def _bootstrap(self, path: str) -> None:
        # Walk recursively through the directory
        for root, _subdirs, files in os.walk(path):
            for fname in files:
                basename = os.path.basename(fname)
                if basename in self._pathmap:
                    # Verify that the folder does not contain multiple files of the same name.
                    raise ValueError("More than one {} in folder".format(basename))
                # Index paths by their basename.
                self._pathmap[basename] = os.path.join(root, fname)
                # Build a lock for each file to synchronize reads.
                self._locks[basename] = RLock()

    def _read_csv(self, filename: str) -> pd.DataFrame:
        path = self._pathmap.get(filename)
        columns = self._config.nodes.get(filename, {}).get("required_columns", [])

        if path is None or os.path.getsize(path) == 0:
            # The file is missing or empty. Return an empty
            # DataFrame containing any required columns.
            return empty_df(columns)

        # If the file isn't in the zip, return an empty DataFrame.
        with open(path, "rb") as f:
            encoding = detect_encoding(f)

        df = pd.read_csv(path, dtype=np.unicode, encoding=encoding, index_col=False)

        # Strip leading/trailing whitespace from column names
        df.rename(columns=lambda x: x.strip(), inplace=True)

        if not df.empty:
            # Strip leading/trailing whitespace from column values
            for col in df.columns:
                df[col] = df[col].str.strip()

        return df

    def _filter(self, filename: str, df: pd.DataFrame) -> pd.DataFrame:
        """Apply view filters"""
        view = self._view.get(filename)
        if view is None:
            return df

        for col, values in view.items():
            # If applicable, filter this dataframe by the given set of values
            if col in df.columns:
                df = df[df[col].isin(setwrap(values))]

        return df

    def _prune(self, filename: str, df: pd.DataFrame) -> pd.DataFrame:
        """Depth-first search through the dependency graph
        and prune dependent DataFrames along the way.
        """
        dependencies = []
        for _, depf, data in self._config.out_edges(filename, data=True):
            deps = data.get("dependencies")
            if deps is None:
                msg = f"Edge missing `dependencies` attribute: {filename}->{depf}"
                raise ValueError(msg)
            dependencies.append((depf, deps))

        if not dependencies:
            return df

        for depfile, column_pairs in dependencies:
            # Read the filtered, cached file dependency
            depdf = self.get(depfile)
            for deps in column_pairs:
                col = deps[filename]
                depcol = deps[depfile]
                # If applicable, prune this dataframe by the other
                if col in df.columns and depcol in depdf.columns:
                    df = df[df[col].isin(depdf[depcol])]

        return df

    def _convert_types(self, filename: str, df: pd.DataFrame) -> None:
        """
        Apply type conversions
        """
        if df.empty:
            return

        converters = self._config.nodes.get(filename, {}).get("converters", {})
        for col, converter in converters.items():
            if col in df.columns:
                df[col] = converter(df[col])

    def _transform(self, filename: str, df: pd.DataFrame) -> pd.DataFrame:
        transformations = self._config.nodes.get(filename, {}).get(
            "transformations", []
        )

        for transform in transformations:
            df = transform(df)

        return df



from .__version__ import __version__
from .readers import (
    load_feed,
    load_geo_feed,
    load_raw_feed,
    read_busiest_date,
    read_busiest_week,
    read_service_ids_by_date,
    read_dates_by_service_ids,
    read_trip_counts_by_date,
)
from .writers import extract_feed


__all__ = [
    "__version__",
    "extract_feed",
    "load_feed",
    "load_geo_feed",
    "load_raw_feed",
    "read_busiest_date",
    "read_busiest_week",
    "read_service_ids_by_date",
    "read_dates_by_service_ids",
    "read_trip_counts_by_date",
]



from collections import defaultdict
import datetime
import os
import shutil
import tempfile
from typing import DefaultDict, Dict, FrozenSet, Optional, Set, Tuple
import weakref

from isoweek import Week
import networkx as nx

from .config import default_config, geo_config, empty_config, reroot_graph
from .gtfs import Feed
from .parsers import vparse_date
from .types import View
from .utilities import remove_node_attributes


DAY_NAMES = (
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
)


def load_feed(
    path: str, view: Optional[View] = None, config: Optional[nx.DiGraph] = None
) -> Feed:
    config = default_config() if config is None else config
    view = {} if view is None else view

    if not nx.is_directed_acyclic_graph(config):
        raise ValueError("Config must be a DAG")

    if os.path.isdir(path):
        feed = _load_feed(path, view, config)
    elif os.path.isfile(path):
        feed = _unpack_feed(path, view, config)
    else:
        raise ValueError("File or path not found: {}".format(path))

    return feed


def load_raw_feed(path: str) -> Feed:
    return load_feed(path, view={}, config=empty_config())


def load_geo_feed(path: str, view: Optional[View] = None) -> Feed:
    return load_feed(path, view=view, config=geo_config())


def read_busiest_date(path: str) -> Tuple[datetime.date, FrozenSet[str]]:
    """Find the earliest date with the most trips"""
    feed = load_raw_feed(path)
    return _busiest_date(feed)


def read_busiest_week(path: str) -> Dict[datetime.date, FrozenSet[str]]:
    """Find the earliest week with the most trips"""
    feed = load_raw_feed(path)
    return _busiest_week(feed)


def read_service_ids_by_date(path: str) -> Dict[datetime.date, FrozenSet[str]]:
    """Find all service identifiers by date"""
    feed = load_raw_feed(path)
    return _service_ids_by_date(feed)


def read_dates_by_service_ids(
    path: str
) -> Dict[FrozenSet[str], FrozenSet[datetime.date]]:
    """Find dates with identical service"""
    feed = load_raw_feed(path)
    return _dates_by_service_ids(feed)


def read_trip_counts_by_date(path: str) -> Dict[datetime.date, int]:
    """A useful proxy for busyness"""
    feed = load_raw_feed(path)
    return _trip_counts_by_date(feed)


def _unpack_feed(path: str, view: View, config: nx.DiGraph) -> Feed:
    tmpdir = tempfile.mkdtemp()
    shutil.unpack_archive(path, tmpdir)
    feed: Feed = _load_feed(tmpdir, view, config)

    # Eager cleanup
    feed._delete_after_reading = True

    def finalize() -> None:
        shutil.rmtree(tmpdir)

    # Lazy cleanup
    weakref.finalize(feed, finalize)

    return feed


def _load_feed(path: str, view: View, config: nx.DiGraph) -> Feed:
    """Multi-file feed filtering"""
    config_ = remove_node_attributes(config, ["converters", "transformations"])
    feed_ = Feed(path, view={}, config=config_)
    for filename, column_filters in view.items():
        config_ = reroot_graph(config_, filename)
        view_ = {filename: column_filters}
        feed_ = Feed(feed_, view=view_, config=config_)
    return Feed(feed_, config=config)


def _busiest_date(feed: Feed) -> Tuple[datetime.date, FrozenSet[str]]:
    service_ids_by_date = _service_ids_by_date(feed)
    trip_counts_by_date = _trip_counts_by_date(feed)

    def max_by(kv: Tuple[datetime.date, int]) -> Tuple[int, int]:
        date, count = kv
        return count, -date.toordinal()

    date, _ = max(trip_counts_by_date.items(), key=max_by)
    service_ids = service_ids_by_date[date]

    return date, service_ids


def _busiest_week(feed: Feed) -> Dict[datetime.date, FrozenSet[str]]:
    service_ids_by_date = _service_ids_by_date(feed)
    trip_counts_by_date = _trip_counts_by_date(feed)

    weekly_trip_counts: DefaultDict[Week, int] = defaultdict(int)
    weekly_dates: DefaultDict[Week, Set[datetime.date]] = defaultdict(set)
    for date in service_ids_by_date.keys():
        key = Week.withdate(date)
        weekly_trip_counts[key] += trip_counts_by_date[date]
        weekly_dates[key].add(date)

    def max_by(kv: Tuple[Week, int]) -> Tuple[int, int]:
        week, count = kv
        return count, -week.toordinal()

    week, _ = max(weekly_trip_counts.items(), key=max_by)
    dates = sorted(weekly_dates[week])

    return {date: service_ids_by_date[date] for date in dates}


def _service_ids_by_date(feed: Feed) -> Dict[datetime.date, FrozenSet[str]]:
    results: DefaultDict[datetime.date, Set[str]] = defaultdict(set)
    removals: DefaultDict[datetime.date, Set[str]] = defaultdict(set)

    service_ids = set(feed.trips.service_id)
    calendar = feed.calendar
    caldates = feed.calendar_dates

    if not calendar.empty:
        # Only consider calendar.txt rows with applicable trips
        calendar = calendar[calendar.service_id.isin(service_ids)].copy()

    if not caldates.empty:
        # Only consider calendar_dates.txt rows with applicable trips
        caldates = caldates[caldates.service_id.isin(service_ids)].copy()

    if not calendar.empty:
        # Parse dates
        calendar.start_date = vparse_date(calendar.start_date)
        calendar.end_date = vparse_date(calendar.end_date)

        # Build up results dict from calendar ranges
        for _, cal in calendar.iterrows():
            start = cal.start_date.toordinal()
            end = cal.end_date.toordinal()

            dow = {i: cal[day] for i, day in enumerate(DAY_NAMES)}
            for ordinal in range(start, end + 1):
                date = datetime.date.fromordinal(ordinal)
                if int(dow[date.weekday()]):
                    results[date].add(cal.service_id)

    if not caldates.empty:
        # Parse dates
        caldates.date = vparse_date(caldates.date)

        # Split out additions and removals
        cdadd = caldates[caldates.exception_type == "1"]
        cdrem = caldates[caldates.exception_type == "2"]

        # Add to results by date
        for _, cd in cdadd.iterrows():
            results[cd.date].add(cd.service_id)

        # Collect removals
        for _, cd in cdrem.iterrows():
            removals[cd.date].add(cd.service_id)

        # Finally, process removals by date
        for date in removals:
            for service_id in removals[date]:
                if service_id in results[date]:
                    results[date].remove(service_id)

            # Drop the key from results if no service present
            if len(results[date]) == 0:
                del results[date]

    assert results, "No service found in feed."

    return {k: frozenset(v) for k, v in results.items()}


def _dates_by_service_ids(feed: Feed) -> Dict[FrozenSet[str], FrozenSet[datetime.date]]:
    results: DefaultDict[FrozenSet[str], Set[datetime.date]] = defaultdict(set)
    for date, service_ids in _service_ids_by_date(feed).items():
        results[service_ids].add(date)
    return {k: frozenset(v) for k, v in results.items()}


def _trip_counts_by_date(feed: Feed) -> Dict[datetime.date, int]:
    results: DefaultDict[datetime.date, int] = defaultdict(int)
    trips = feed.trips
    for service_ids, dates in _dates_by_service_ids(feed).items():
        trip_count = trips[trips.service_id.isin(service_ids)].shape[0]
        for date in dates:
            results[date] += trip_count
    return dict(results)



# flake8: noqa E501

import networkx as nx
import pandas as pd

from .parsers import vparse_date, vparse_time


def empty_config() -> nx.DiGraph:
    return nx.DiGraph()


def default_config() -> nx.DiGraph:
    G = empty_config()
    add_edge_config(G)
    add_node_config(G)
    return G


def geo_config() -> nx.DiGraph:
    G = default_config()
    add_geo_config(G)
    return G


def add_edge_config(g: nx.DiGraph) -> nx.DiGraph:
    g.add_edges_from(
        [
            (
                "agency.txt",
                "routes.txt",
                {
                    "dependencies": [
                        {"agency.txt": "agency_id", "routes.txt": "agency_id"}
                    ]
                },
            ),
            (
                "calendar.txt",
                "trips.txt",
                {
                    "dependencies": [
                        {"calendar.txt": "service_id", "trips.txt": "service_id"}
                    ]
                },
            ),
            (
                "calendar_dates.txt",
                "trips.txt",
                {
                    "dependencies": [
                        {"calendar_dates.txt": "service_id", "trips.txt": "service_id"}
                    ]
                },
            ),
            (
                "fare_attributes.txt",
                "fare_rules.txt",
                {
                    "dependencies": [
                        {"fare_attributes.txt": "fare_id", "fare_rules.txt": "fare_id"}
                    ]
                },
            ),
            (
                "fare_rules.txt",
                "stops.txt",
                {
                    "dependencies": [
                        {"fare_rules.txt": "origin_id", "stops.txt": "zone_id"},
                        {"fare_rules.txt": "destination_id", "stops.txt": "zone_id"},
                        {"fare_rules.txt": "contains_id", "stops.txt": "zone_id"},
                    ]
                },
            ),
            (
                "fare_rules.txt",
                "routes.txt",
                {
                    "dependencies": [
                        {"fare_rules.txt": "route_id", "routes.txt": "route_id"}
                    ]
                },
            ),
            (
                "frequencies.txt",
                "trips.txt",
                {
                    "dependencies": [
                        {"frequencies.txt": "trip_id", "trips.txt": "trip_id"}
                    ]
                },
            ),
            (
                "routes.txt",
                "trips.txt",
                {"dependencies": [{"routes.txt": "route_id", "trips.txt": "route_id"}]},
            ),
            (
                "shapes.txt",
                "trips.txt",
                {"dependencies": [{"shapes.txt": "shape_id", "trips.txt": "shape_id"}]},
            ),
            (
                "stops.txt",
                "stop_times.txt",
                {
                    "dependencies": [
                        {"stops.txt": "stop_id", "stop_times.txt": "stop_id"}
                    ]
                },
            ),
            (
                "stop_times.txt",
                "trips.txt",
                {
                    "dependencies": [
                        {"stop_times.txt": "trip_id", "trips.txt": "trip_id"}
                    ]
                },
            ),
            (
                "transfers.txt",
                "stops.txt",
                {
                    "dependencies": [
                        {"transfers.txt": "from_stop_id", "stops.txt": "stop_id"},
                        {"transfers.txt": "to_stop_id", "stops.txt": "stop_id"},
                    ]
                },
            ),
        ]
    )


def add_node_config(g: nx.DiGraph) -> nx.DiGraph:
    g.add_nodes_from(
        [
            (
                "agency.txt",
                {"required_columns": ("agency_name", "agency_url", "agency_timezone")},
            ),
            (
                "calendar.txt",
                {
                    "converters": {
                        "start_date": vparse_date,
                        "end_date": vparse_date,
                        "monday": pd.to_numeric,
                        "tuesday": pd.to_numeric,
                        "wednesday": pd.to_numeric,
                        "thursday": pd.to_numeric,
                        "friday": pd.to_numeric,
                        "saturday": pd.to_numeric,
                        "sunday": pd.to_numeric,
                    },
                    "required_columns": (
                        "service_id",
                        "monday",
                        "tuesday",
                        "wednesday",
                        "thursday",
                        "friday",
                        "saturday",
                        "sunday",
                        "start_date",
                        "end_date",
                    ),
                },
            ),
            (
                "calendar_dates.txt",
                {
                    "converters": {
                        "date": vparse_date,
                        "exception_type": pd.to_numeric,
                    },
                    "required_columns": ("service_id", "date", "exception_type"),
                },
            ),
            (
                "fare_attributes.txt",
                {
                    "converters": {
                        "price": pd.to_numeric,
                        "payment_method": pd.to_numeric,
                        "transfer_duration": pd.to_numeric,
                    },
                    "required_columns": (
                        "fare_id",
                        "price",
                        "currency_type",
                        "payment_method",
                        "transfers",
                    ),
                },
            ),
            ("fare_rules.txt", {"required_columns": ("fare_id",)}),
            (
                "feed_info.txt",
                {
                    "converters": {
                        "feed_start_date": vparse_date,
                        "feed_end_date": vparse_date,
                    },
                    "required_columns": (
                        "feed_publisher_name",
                        "feed_publisher_url",
                        "feed_lang",
                    ),
                },
            ),
            (
                "frequencies.txt",
                {
                    "converters": {
                        "headway_secs": pd.to_numeric,
                        "exact_times": pd.to_numeric,
                        "start_time": vparse_time,
                        "end_time": vparse_time,
                    },
                    "required_columns": (
                        "trip_id",
                        "start_time",
                        "end_time",
                        "headway_secs",
                    ),
                },
            ),
            (
                "routes.txt",
                {
                    "converters": {"route_type": pd.to_numeric},
                    "required_columns": (
                        "route_id",
                        "route_short_name",
                        "route_long_name",
                        "route_type",
                    ),
                },
            ),
            (
                "shapes.txt",
                {
                    "converters": {
                        "shape_pt_lat": pd.to_numeric,
                        "shape_pt_lon": pd.to_numeric,
                        "shape_pt_sequence": pd.to_numeric,
                        "shape_dist_traveled": pd.to_numeric,
                    },
                    "required_columns": (
                        "shape_id",
                        "shape_pt_lat",
                        "shape_pt_lon",
                        "shape_pt_sequence",
                    ),
                },
            ),
            (
                "stops.txt",
                {
                    "converters": {
                        "stop_lat": pd.to_numeric,
                        "stop_lon": pd.to_numeric,
                        "location_type": pd.to_numeric,
                        "wheelchair_boarding": pd.to_numeric,
                        "pickup_type": pd.to_numeric,
                        "drop_off_type": pd.to_numeric,
                        "shape_dist_traveled": pd.to_numeric,
                        "timepoint": pd.to_numeric,
                    },
                    "required_columns": (
                        "stop_id",
                        "stop_name",
                        "stop_lat",
                        "stop_lon",
                    ),
                },
            ),
            (
                "stop_times.txt",
                {
                    "converters": {
                        "arrival_time": vparse_time,
                        "departure_time": vparse_time,
                        "pickup_type": pd.to_numeric,
                        "shape_dist_traveled": pd.to_numeric,
                        "stop_sequence": pd.to_numeric,
                        "timepoint": pd.to_numeric,
                    },
                    "required_columns": (
                        "trip_id",
                        "arrival_time",
                        "departure_time",
                        "stop_id",
                        "stop_sequence",
                    ),
                },
            ),
            (
                "transfers.txt",
                {
                    "converters": {
                        "transfer_type": pd.to_numeric,
                        "min_transfer_time": pd.to_numeric,
                    },
                    "required_columns": ("from_stop_id", "to_stop_id", "transfer_type"),
                },
            ),
            (
                "trips.txt",
                {
                    "converters": {
                        "direction_id": pd.to_numeric,
                        "wheelchair_accessible": pd.to_numeric,
                        "bikes_allowed": pd.to_numeric,
                    },
                    "required_columns": ("route_id", "service_id", "trip_id"),
                },
            ),
        ]
    )


def add_geo_config(g: nx.DiGraph) -> nx.DiGraph:
    from .geo import build_shapes, build_stops

    for node, transform in (("shapes.txt", build_shapes), ("stops.txt", build_stops)):
        assert node in g.nodes, "Missing config for node: {}".format(node)

        if "transformations" not in g.nodes[node]:
            g.nodes[node]["transformations"] = []

        g.nodes[node]["transformations"].append(transform)

    return g


def reroot_graph(G: nx.DiGraph, node: str) -> nx.DiGraph:
    """Return a copy of the graph rooted at the given node"""
    G = G.copy()
    for n, successors in list(nx.bfs_successors(G, source=node)):
        for s in successors:
            G.add_edge(s, n, **G.edges[n, s])
            G.remove_edge(n, s)
    return G



import os
import shutil
import tempfile
from multiprocessing.pool import ThreadPool
from typing import Collection, Optional

import networkx as nx

from .config import default_config
from .gtfs import Feed
from .readers import load_feed
from .types import View
from .utilities import remove_node_attributes


DEFAULT_NODES = frozenset(default_config().nodes())


def extract_feed(
    inpath: str, outpath: str, view: View, config: nx.DiGraph = None
) -> str:
    """Extract a subset of a GTFS zip into a new file"""
    config = default_config() if config is None else config
    config = remove_node_attributes(config, "converters")
    feed = load_feed(inpath, view, config)
    return write_feed_dangerously(feed, outpath)


def write_feed_dangerously(
    feed: Feed, outpath: str, nodes: Optional[Collection[str]] = None
) -> str:
    """Naively write a feed to a zipfile

    This function provides no sanity checks. Use it at
    your own risk.
    """
    nodes = DEFAULT_NODES if nodes is None else nodes
    with tempfile.TemporaryDirectory() as tmpdir:

        def write_node(node):
            df = feed.get(node)
            if not df.empty:
                path = os.path.join(tmpdir, node)
                df.to_csv(path, index=False)

        pool = ThreadPool(len(nodes))
        try:
            pool.map(write_node, nodes)
        finally:
            pool.terminate()

        if outpath.endswith(".zip"):
            outpath, _ = os.path.splitext(outpath)

        outpath = shutil.make_archive(outpath, "zip", tmpdir)

    return outpath



from typing import Dict, List

import pandas as pd

try:
    import geopandas as gpd
    from shapely.geometry import LineString, Point
except ImportError as impexc:
    print(impexc)
    print("You must install GeoPandas to use this module.")
    raise


DEFAULT_CRS = {"init": "EPSG:4326"}


def build_shapes(df: pd.DataFrame) -> gpd.GeoDataFrame:
    if df.empty:
        return gpd.GeoDataFrame({"shape_id": [], "geometry": []}, crs=DEFAULT_CRS)

    data: Dict[str, List] = {"shape_id": [], "geometry": []}
    for shape_id, shape in df.sort_values("shape_pt_sequence").groupby("shape_id"):
        data["shape_id"].append(shape_id)
        data["geometry"].append(
            LineString(list(zip(shape.shape_pt_lon, shape.shape_pt_lat)))
        )

    return gpd.GeoDataFrame(data, crs=DEFAULT_CRS)


def build_stops(df: pd.DataFrame) -> gpd.GeoDataFrame:
    if df.empty:
        return gpd.GeoDataFrame(df, geometry=[], crs=DEFAULT_CRS)

    df["geometry"] = df.apply(lambda s: Point(s.stop_lon, s.stop_lat), axis=1)

    df.drop(["stop_lon", "stop_lat"], axis=1, inplace=True)

    return gpd.GeoDataFrame(df, crs={"init": "EPSG:4326"})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import io

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

# About dict to store version and package info
about = dict()
with io.open("partridge/__version__.py", "r", encoding="utf-8") as f:
    exec(f.read(), about)

requirements = [
    "cchardet",
    'functools32;python_version<"3"',
    "networkx>=2.0",
    "pandas",
    "isoweek",
]

setup_requirements = ["pytest-runner"]

test_requirements = ["black", "flake8", "pytest"]

setup(
    name="partridge",
    version=about["__version__"],
    description="Partridge is a python library for working with GTFS "
    "feeds using pandas DataFrames.",
    long_description=readme + "\n\n" + history,
    long_description_content_type='text/x-rst',
    author="Danny Whalen",
    author_email="daniel.r.whalen@gmail.com",
    url="https://github.com/remix/partridge",
    packages=find_packages(include=["partridge"]),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords="partridge",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    python_requires=">=3.6, <4",
    test_suite="tests",
    tests_require=test_requirements,
    setup_requires=setup_requirements,
    extras_require={"full": ["geopandas"]},
)



import datetime
from functools import lru_cache
import numpy as np

DATE_FORMAT = "%Y%m%d"


# Why 2^17? See https://git.io/vxB2P.
@lru_cache(maxsize=2 ** 17)
def parse_time(val: str) -> np.float64:
    if val is np.nan:
        return val

    val = val.strip()

    if val == "":
        return np.nan

    h, m, s = val.split(":")
    ssm = int(h) * 3600 + int(m) * 60 + int(s)

    # pandas doesn't have a NaN int, use floats
    return np.float64(ssm)


def parse_date(val: str) -> datetime.date:
    return datetime.datetime.strptime(val, DATE_FORMAT).date()


vparse_date = np.vectorize(parse_date)
vparse_time = np.vectorize(parse_time)
